/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/                  

#include "Quakedef.h"

// --------------------------------------------------------------
#define DECAL_GLOWTIME 60
#define DECAL_HASHSIZE 0x00008000
#define DECAL_HASHMASK 0x00007FFF

typedef struct decal_s
{
	vec3_t			org;
	vec3_t			direction;
	vec3_t			verts[64]; // vertices
	vec_t[2]		st[64]; // texcoords
	int				numVerts;
	float			color[4];
	image_t			*tex;
	msurface_t		*surface;
	byte			*lightmapaddress;
	int				lightmapstep;
	unsigned int	hash;
	int				type;
	float			TimeOfCreation;
	struct decal_s* next; // points to the (global) decal that was created right after this one
#ifdef _DEBUG
	vec3_t			debugBounds[64];
	int				numBounds;
#endif
}
decal_t;

decal_t *decals_normal;    // allocated memory for all decals
decal_t *decals_small;  
unsigned int decal_hash[DECAL_HASHSIZE];

static int decal_capacity;

cvar_t *gl_decals;
cvar_t *gl_decals_max;
cvar_t *gl_decals_viewdistance;
cvar_t *gl_decals_lighting;

int decal_small_count;
decal_t *decal_small_oldest; // first 
decal_t *decal_small_newest; // last
decal_t *decal_small_free;  // first free

int decal_normal_count;
decal_t *decal_normal_oldest;
decal_t *decal_normal_newest; 
decal_t *decal_normal_free;  


void decal_normal_push(decal_t *freeOne) {
  freeOne->next = decal_normal_free;
  decal_normal_free = freeOne;
}

decal_t* decal_normal_pop(void) 
{
  if (decal_normal_free) 
  {
    decal_normal_newest->next = decal_normal_free;    
    decal_normal_newest = decal_normal_newest->next;    
    decal_normal_free = decal_normal_free->next;
    if(!decal_normal_oldest) decal_normal_oldest = decal_normal_newest;
  }else if (decal_normal_count < decal_capacity) {
    decal_normal_newest->next = decals_normal + decal_normal_count++;
    decal_normal_newest = decal_normal_newest->next;
    if(!decal_normal_oldest) decal_normal_oldest = decal_normal_newest;
  }else {
    decal_normal_newest->next = decal_normal_oldest;
    decal_normal_oldest = decal_normal_oldest->next;
    decal_normal_newest = decal_normal_newest->next;    
  }
  decal_normal_newest->next=NULL;
  return decal_normal_newest;
}

void decal_small_push(decal_t *freeOne) {
  freeOne->next = decal_small_free;
  decal_small_free = freeOne;
}

decal_t* decal_small_pop(void) {
  if (decal_small_free) {
    decal_small_newest->next = decal_small_free;    
    decal_small_newest = decal_small_newest->next;    
    decal_small_free = decal_small_free->next;
    if(!decal_small_oldest) decal_small_oldest = decal_small_newest;
  }else if (decal_small_count < decal_capacity) {
    decal_small_newest->next = decals_small + decal_small_count++;
    decal_small_newest = decal_small_newest->next;
    if(!decal_small_oldest) decal_small_oldest = decal_small_newest;
  }else {
    decal_small_newest->next = decal_small_oldest;
    decal_small_oldest = decal_small_oldest->next;
    decal_small_newest = decal_small_newest->next;    
  }
  decal_small_newest->next=NULL;
  return decal_small_newest;
}


void R_Decals_Start(void)
{
	decal_capacity = gl_decals->value;

	if(decal_capacity<=0){
		decal_capacity=0;
		return;
	}
	decal_capacity = (decal_capacity+1) >> 1;

	// force size 32 at least
	if(decal_capacity<32) {
	   decal_capacity=32;
	}
	decals_normal = (decal_t *) Q_malloc((decal_capacity <<1) * sizeof(decal_t));
	decals_small = decals_normal + decal_capacity;    
	R_Decals_Newmap();  
}

void R_Decals_Shutdown(void)
{
	if(!decal_capacity) return;
	free(decals_normal);  
}

void R_Decals_Newmap(void)
{
	if(!decal_capacity) return;  
	memset(decals_normal, 0, (decal_capacity <<1) * sizeof(decal_t));
	memset(decal_hash, 0, DECAL_HASHSIZE* sizeof(int));

	decal_normal_newest = decals_normal; //point to any free memory
	decal_small_newest = decals_small; //point to any free memory

	decal_normal_oldest = NULL;//decals_normal; // will initially point to the first decal
	decal_small_oldest = NULL;//decals_small; 

	decal_normal_count = 0;
	decal_small_count = 0;

	decal_normal_free = NULL;
	decal_small_free = NULL;
}

void R_Decals_Init(void)
{
	gl_decals = Cvar_Get( "gl_decals", "1024", CVAR_ARCHIVE );
	gl_decals_max = Cvar_Get( "gl_decals_max", "400", CVAR_ARCHIVE );
	gl_decals_lighting = Cvar_Get( "gl_decals_lighting", "1", CVAR_ARCHIVE );
	gl_decals_viewdistance = Cvar_Get( "gl_decals_viewdistance", "1280", CVAR_ARCHIVE );
}

// these are static globals only to avoid putting unnecessary things on the stack
static vec3_t decalorg;
static float decalbestdist[64];
static msurface_t *decalbestsurf[64];
static int decalbestlightmapofs[64];
static int bestdecalnum;
static int bestdecal;

void R_RecursiveDecalSurface (mnode_t *node)
{
	// these are static because only one occurance of them need exist at once, so avoid putting them on the stack
	static float ndist, dist;
	static msurface_t *surf, *endsurf;
	static vec3_t impact;
	static int ds, dt;

loc0:
  	if (node->contents >= 0) return;

	ndist = PlaneDiff(decalorg, node->plane);  
	
	if (ndist > 32)
	{
		node = node->children[0];
		goto loc0;
	}

	if (ndist < -32)
	{
		node = node->children[1];
		goto loc0;
	}  

	// mark the polygons
	surf = r_worldmodel->surfaces + node->firstsurface;
	endsurf = surf + node->numsurfaces;

	for (surf=surf;surf < endsurf;surf++)
	{
		if (surf->texinfo->flags & (SURF_DRAWSKY|SURF_DRAWWATER|SURF_DRAWLAVA|SURF_DRAWSLIME|SURF_DRAWTELE|SURF_DRAWUNDERWATER)) continue;

  		dist = PlaneDiff(decalorg, surf->plane);

		if (surf->flags & SURF_PLANEBACK)
			dist = -dist;
		if (dist < 0)
			continue;
		
		impact[0] = decalorg[0] - surf->plane->normal[0] * dist;
		impact[1] = decalorg[1] - surf->plane->normal[1] * dist;
		impact[2] = decalorg[2] - surf->plane->normal[2] * dist;

		ds = (int) (DotProduct(impact, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3]);
		dt = (int) (DotProduct(impact, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3]);

		if (ds < surf->texturemins[0] || dt < surf->texturemins[1])
			continue;
		
		ds -= surf->texturemins[0];
		dt -= surf->texturemins[1];
		
		if (ds > surf->extents[0] || dt > surf->extents[1])
			continue;
    
		if (dist < decalbestdist[bestdecal]) {
			bestdecal = bestdecalnum;
		}
		decalbestsurf[bestdecalnum] = surf;
		decalbestdist[bestdecalnum] = dist;
		decalbestlightmapofs[bestdecalnum] = (dt >> 4) * ((surf->extents[0] >> 4) + 1) + (ds >> 4);
		bestdecalnum++;
	}

	if (node->children[0]->contents < 0)
	{
		if (node->children[1]->contents < 0)
		{
			R_RecursiveDecalSurface (node->children[0]);
			node = node->children[1];
			goto loc0;
		}
		else
		{
			node = node->children[0];
			goto loc0;
		}
	}
	else if (node->children[1]->contents < 0)
	{
		node = node->children[1];
		goto loc0;
	}
}

// decal clipping
static int SplitLine(vec3_t v1,vec3_t v2,vec3_t s, cplane_t p) {
	float d1,d2,d;
	  
	d1=DotProduct(v1,p.normal)-p.dist;
	d2=DotProduct(v2,p.normal)-p.dist;
	  
	if (d1>0 && d2>0) return 2;
	if (d1<0 && d2<0) return -2;
	if (d1-d2 == 0) return 0;
	  
	d=d1/(d1-d2);
	s[0]=v1[0]+(v2[0]-v1[0])*d;
	s[1]=v1[1]+(v2[1]-v1[1])*d;
	s[2]=v1[2]+(v2[2]-v1[2])*d;   

	if(d1>d2) {
		return 1;
	}
	return -1;
}

static int CutPolyWithPlane(vec3_t *verts,int numVerts,cplane_t plane) {  
	vec3_t		NewVerts[64];
	vec3_t		v0;
	int			numNewVerts;
	int			i,k;	
	int			f;
	qboolean	dirty=false;

	numNewVerts=0;  

	for(i=0;i<numVerts;i++) {
		k = i+1; 
		if(k>=numVerts) k=0;
		f = SplitLine(verts[i],verts[k],v0,plane);
	    
		if(f == 2 || f==0) continue;  

		if(f == -2) {  
			// both points are inside
			VectorCopy(verts[i],NewVerts[numNewVerts]);
			if(numNewVerts >= 63) return 0; // discard decal
			numNewVerts++;
		} else if(f == 1 ) { 
			// i is outside
			dirty=true;
			VectorCopy(v0,NewVerts[numNewVerts]);    
			if(numNewVerts >= 63) return 0;
			numNewVerts++;
		} else if(f == -1 ) { 
			// k is outside
			dirty=true;
			VectorCopy(verts[i],NewVerts[numNewVerts]);
			numNewVerts++;
			VectorCopy(v0,NewVerts[numNewVerts]);
			if(numNewVerts >= 63) return 0;
			numNewVerts++;
		}
	}

	if(dirty) {
		for(i=0;i<numNewVerts;i++) {
			VectorCopy(NewVerts[i],verts[i]);
		}    
		return numNewVerts;
	} else {
		return numVerts;
	}
}

void decal_CutToEdges (msurface_t *fa, decal_t *D,const vec3_t tex_0,const vec3_t tex_s,const vec3_t tex_t)
{	
	vec3_t		verts[64];
	 vec3_t		v0;
	int			numverts;
	int			i,n0,n1;
	int			lindex;
	float		*vec;

	// convert edges back to a normal polygon	
	numverts = 0;

	for (i=0 ; i<fa->numedges && i<64 ; i++)
	{
		lindex = r_worldmodel->surfedges[fa->firstedge + i];

		if (lindex > 0)
			vec = r_worldmodel->vertexes[r_worldmodel->edges[lindex].v[0]].position;
		else
			vec = r_worldmodel->vertexes[r_worldmodel->edges[-lindex].v[1]].position;
		VectorCopy (vec, verts[numverts]);
		numverts++;
	}

	for(i=0;i<numverts;i++){
		cplane_t plane;

		n0 = i;
		n1 = i+1;

		if(n1>=numverts) n1 = 0;

		// create plane from edge
		VectorSubtract(verts[n1],verts[n0],v0);
		CrossProduct(v0,D->direction,plane.normal);
		VectorNormalize(plane.normal);

		//VectorNegate(plane.normal,plane.normal);

		plane.dist = DotProduct(plane.normal,verts[n0]);

		// cut decal poly
		D->numVerts = CutPolyWithPlane(D->verts,D->numVerts,plane);
		if(D->numVerts==0) return;
	}  

	// create texture coords
	for(i=0;i<D->numVerts;i++){    
		D->st[i][0] = DotProduct((D->verts[i]),tex_s) - tex_0[0];
		D->st[i][1] = DotProduct((D->verts[i]),tex_t) - tex_0[1];
	}

#ifdef _DEBUG 
	memcpy(D->debugBounds,verts,numverts*sizeof(vec3_t));
	D->numBounds = numverts;
#endif
}

//sul - checks if org does already exist    // returns 0 if decal can't be set
//__inline
unsigned int decal_duplicateOrg(vec3_t org) {
	int i;  
	int x = ((int)(4096 + org[0]) >> 4) & 0x3ff;
	int y = ((int)(4096 + org[1]) >> 4) & 0x3ff;
	int z = ((int)(4096 + org[2]) >> 4) & 0x3ff;
	unsigned int hash= (z<<20) + (y<<10) + x;
	unsigned int hash_index = hash % DECAL_HASHSIZE;

	for(i=0;i<4;i++) {
		if (decal_hash[hash_index+i] == 0) {
			decal_hash[hash_index+i] = hash;
			return hash_index+i; // return position in array so that it can clear itself later
		}
		if (hash == decal_hash[hash_index+i]) {
			return 0; // deny, already exists
		}
	}
	return 0; //deny, doesn't exist but the array spots are already taken
}

void R_Decal(vec3_t org, char *tex, float scale, float cred, float cgreen, float cblue, float alpha, int decal_type )
{
	vec3_t			center, right, up,tex_0;
	decal_t			*decal;
	unsigned int	hash;
	int				i;
	float			angle,f;

	if (!decal_capacity) return;

	if(decal_type & DECAL_HASH) 
	{
		hash = decal_duplicateOrg(org); 
		if (!hash) return;
	} 
	else 
	{
		hash = 0;
	}
	// find the best surface to place the decal on
	decalbestsurf[0] = NULL;
	decalbestdist[0] = 16;
	decalbestlightmapofs[0] = 0;

	VectorCopy(org, decalorg);

	bestdecal=0;
	bestdecalnum=0;

	R_RecursiveDecalSurface (r_worldmodel->nodes);

	// abort if no suitable surface was found
	if (bestdecalnum == 0) return;

	angle = rand()* (360.0/RAND_MAX);

	for(i=0;i<bestdecalnum;i++) {

		if(i!=bestdecal && 0.5 > DotProduct(decalbestsurf[i]->plane->normal,decalbestsurf[bestdecal]->plane->normal)) continue;
    
		// pop unused or maybe oldest decal
		if(decal_type & DECAL_SMALL){ 
			decal = decal_small_pop();
		} else {
			decal = decal_normal_pop();
		}
	    
		decal_hash[decal->hash] = 0; // this one got cleared;
		decal->hash = hash;
		decal->type = decal_type;
		decal->TimeOfCreation = r_newrefdef.time;
	    
		decal->tex = GL_FindImage(tex,it_wall);
		VectorCopy(decalbestsurf[i]->plane->normal, decal->direction);

		// reverse direction
		if (!(decalbestsurf[i]->flags & SURF_PLANEBACK)) {
			VectorNegate(decal->direction, decal->direction);
		    
			decal->org[0] = center[0] = org[0] + decal->direction[0] * decalbestdist[i];
			decal->org[1] = center[1] = org[1] + decal->direction[1] * decalbestdist[i];
			decal->org[2] = center[2] = org[2] + decal->direction[2] * decalbestdist[i];
			// set up the 4 corners
			f = scale*0.5f;
		    
		    
			decal->numVerts=4;
			VectorVectors(decal->direction, right, up,angle);
			VectorScale(right,f,right);
			VectorScale(up,f,up);
		    
			decal->verts[0][0] = center[0] - right[0]  - up[0]; // texcoords 0 1
			decal->verts[0][1] = center[1] - right[1]  - up[1];
			decal->verts[0][2] = center[2] - right[2]  - up[2];
		    
			decal->verts[1][0] = center[0] - right[0]  + up[0]; // texcoords 0 0
			decal->verts[1][1] = center[1] - right[1]  + up[1];
			decal->verts[1][2] = center[2] - right[2]  + up[2];
		    
			decal->verts[2][0] = center[0] + right[0]  + up[0]; // texcoords 1 0
			decal->verts[2][1] = center[1] + right[1]  + up[1];
			decal->verts[2][2] = center[2] + right[2]  + up[2];
		    
			decal->verts[3][0] = center[0] + right[0]  - up[0] ; // texcoords 1 1
			decal->verts[3][1] = center[1] + right[1]  - up[1] ;
			decal->verts[3][2] = center[2] + right[2]  - up[2] ;
		    
			// CLIP DECAL
			f = 1/(2*f*f);
			VectorScale(right,f,right);
			VectorScale(up,f,up);
			tex_0[0] = DotProduct( decal->verts[1], right);
			tex_0[1] = DotProduct( decal->verts[1], up);
			decal_CutToEdges(decalbestsurf[i],decal,tex_0,right,up);
		}

		if(decal->numVerts<3 || decal->numVerts > 10) { // discard
			decal->type |= DECAL_TEMP;
			decal->TimeOfCreation = 0;
			Com_DPrintf ( S_COLOR_GREEN "WARNING: decal discarded due to strange number of verts: %i\n",decal->numVerts);
		}

#ifdef _DEBUG    // this is probably not required
		VectorSubtract(decal->verts[1],decal->verts[0],right);
		VectorSubtract(decal->verts[2],decal->verts[0],up);
		CrossProduct(right,up,center);

		if(VectorLength(center) > 1.1*scale*scale) {
			// discard
			decal->type |= DECAL_TEMP;
			decal->TimeOfCreation = 0;
			Com_Printf ( S_COLOR_RED "WARNING: decal discarded due to invalid size\n");
		}//-----------------------------
#endif
		// store the color
		decal->color[0] = cred;
		decal->color[1] = cgreen;
		decal->color[2] = cblue;
		decal->color[3] = alpha;
		// store the surface information for lighting
		decal->surface = decalbestsurf[i];
		decal->lightmapstep = ((decalbestsurf[i]->extents[0]>>4)+1) * ((decalbestsurf[i]->extents[1]>>4)+1)*3; // LordHavoc: *3 for colored lighting
		if (decalbestsurf[i]->samples) {
			decal->lightmapaddress = decalbestsurf[i]->samples + decalbestlightmapofs[i] * 3; // LordHavoc: *3 for colored lighitng
		} else {
			decal->lightmapaddress = NULL;
		}
	}
}

void GL_DrawDecals (void)
{
	decal_t		*p, *oldp;
	dlight_t	*dl;
	int			lnum;
	float		fdist, frad;
	int			i, j,n,texnum;
	float		fr, fg, fb, fa, mindist;
	byte		*lightmap;
	vec3_t		v;
	msurface_t	*surf;
	float		decal_viewmax;
	int			maxdecals;

	if (!decal_capacity) return;

	mindist = DotProduct(r_origin, vpn) + 4.0f;    

	GL_EnableMultitexture( false );
	GL_TexEnv(GL_MODULATE);
	GLSTATE_ENABLE_BLEND	  
	GLSTATE_DISABLE_ALPHATEST  
	qglEnable(GL_POLYGON_OFFSET);
	qglPolygonOffset(-1,-2);
	qglDepthMask(GL_FALSE); // disable zbuffer updates

	texnum = -1;

	decal_viewmax = gl_decals_viewdistance->value;
	decal_viewmax *=decal_viewmax;

	for(n=0;n<2;n++) {  
		if(n==0) {
			decal_viewmax *=0.25;
			p = decal_small_oldest; // draw small first
			oldp = NULL;
			maxdecals = gl_decals_max->value/2;
		} else {
			decal_viewmax *=4;
			p = decal_normal_oldest;
			oldp = NULL;
			maxdecals = gl_decals_max->value;
		}

		for(/**/;p; oldp=p , p=p->next) {
			// skip decals on surfaces that aren't visible in this frame
			if (p->surface->visframe != r_framecount) continue;

			// skip and free decal if DECAL_TEMP and time is over
			if (p->type & DECAL_TEMP){
				if(p->TimeOfCreation +DECAL_GLOWTIME < r_newrefdef.time ) {

					if(oldp) {
						oldp->next = p->next;
						if( p->type & DECAL_SMALL) decal_small_push(p);
						else decal_normal_push(p);
						p = oldp;
					} else {
						if( p->type & DECAL_SMALL) {
							decal_small_oldest = p->next;
							decal_small_push(p);
							p=decal_small_oldest;
							if(!p) { decal_small_newest = decals_small; break; }
						} else {
							decal_normal_oldest = p->next;
							decal_normal_push(p);
							p=decal_normal_oldest;
							if(!p) { decal_normal_newest = decals_normal; break; }
						}
					}
					continue;
				}
			}     

			// do not render if the decal is behind the view
			if ( DotProduct(p->org, vpn) < mindist)	continue;    

			// do not render if the view origin is behind the decal
			VectorSubtract(p->org, r_origin, v);
			if (DotProduct(p->direction, v) < 0) continue;

			fa = DotProduct(v,v);
			fa = 2 -  fa/decal_viewmax;

			if( fa < 0.05) continue;		// don't draw decals in far distance

			c_decals++;

			if (c_decals>maxdecals) break;	// stop here to save fps

#ifdef _DEBUG
			if(p->numBounds){ 
				qglDisable(GL_TEXTURE_2D); 
				qglColor3f(1,0,1);          
				qglBegin(GL_LINE_STRIP);

				for(i=0;i<p->numBounds;i++){
					qglVertex3fv(p->debugBounds[i]);
				}
				qglVertex3fv(p->debugBounds[0]);
				qglEnd();
				qglEnable(GL_TEXTURE_2D);

			}
#endif

			// get the surface lighting
			surf = p->surface;
			lightmap = p->lightmapaddress;
			fr = fg = fb = 0.0f;

			if (gl_adv_IR->value && r_newrefdef.rdflags & RDF_IRGOGGLES ) {
				fr = 0.5 * p->color[0] + 0.2 * p->color[2];
				fg = 0.6 * p->color[1];
				fb = 0;
				fa = 0.6 * p->color[3];
			} else {
				if (!(p->type & DECAL_GLOW)) {                         
#ifdef _DEBUG
					if(sul_x1->value) { 
						R_LightColor(p->org,v);
						fr = v[0]*p->color[0];
						fg = v[1]*p->color[1];
						fb = v[2]*p->color[2];    
						fa *= p->color[3];
					} else
#endif
					{
						if (lightmap) {
							vec3_t	scale;
							int		maps;

							for (maps = 0 ; maps < MAXLIGHTMAPS && surf->styles[maps] != 255 ;maps++){
								for (i=0 ; i<3 ; i++) {
									scale[i] = r_newrefdef.lightstyles[surf->styles[maps]].rgb[i];
								}
								fr += lightmap[0] * scale[0] * (1.0/255);
								fg += lightmap[1] * scale[1] * (1.0/255);
								fb += lightmap[2] * scale[2] * (1.0/255);
								lightmap += 3*((surf->extents[0]>>4)+1) * ((surf->extents[1]>>4)+1);
							}
						}    

						if (surf->dlightframe == r_framecount) {
							for (lnum=0 ; lnum<r_newrefdef.num_dlights ; lnum++) {
								if ( !(surf->dlightbits & (1<<lnum) ) )	continue;	// not lit by this light

								dl = &r_newrefdef.dlights[lnum];
								VectorSubtract(p->org, dl->origin, v);
								fdist = DotProduct(v, v) + 4096;
								frad = dl->intensity * dl->intensity;

								if (fdist >= frad) continue;

								frad *= 24/ fdist;
								fr += frad * dl->color[0]/255;
								fg += frad * dl->color[1]/255;
								fb += frad * dl->color[2]/255;
							}
						}

						if(fr>1)fr=1;
						if(fg>1)fg=1;
						if(fb>1)fb=1;
						if(fa>1)fa=1;

						fr *= p->color[0];
						fg *= p->color[1];
						fb *= p->color[2];    
						fa *= p->color[3];
					}
				} else { // DECAL_GLOW
					fr = p->color[0];
					fg = p->color[1];
					fb = p->color[2];    
					fa = p->color[3];      
				}
			}

			if (p->type & DECAL_TEMP) {
				fa *= ((DECAL_GLOWTIME + p->TimeOfCreation-r_newrefdef.time)/DECAL_GLOWTIME);    
			}
			qglColor4f(fr,fg,fb,fa);

			j = p->tex->texnum;
			if (texnum != j) {      
				texnum = j;
				GL_Bind(texnum);
			}
			qglBegin(GL_TRIANGLE_FAN);

			for(i=0;i<p->numVerts;i++){
				qglTexCoord2fv(p->st[i]);
				qglVertex3fv(p->verts[i]);
			}
			qglEnd();   
		}

		if (GL_MsgGLError("BLEH")){
			GL_MsgGLError("BLEH");
		}

	}   
	qglDisable(GL_POLYGON_OFFSET);
	qglDepthMask(GL_TRUE); // enable zbuffer updates
	GLSTATE_DISABLE_BLEND	  
	GLSTATE_ENABLE_ALPHATEST
}